<?php
/**
 * @package DBCAPI
 */

class DeathByCaptcha_Exception extends Exception
{}


class DeathByCaptcha_RuntimeException extends DeathByCaptcha_Exception
{}


class DeathByCaptcha_ServerException extends DeathByCaptcha_Exception
{}


class DeathByCaptcha_ClientException extends DeathByCaptcha_Exception
{}


class DeathByCaptcha_InvalidAccountException extends DeathByCaptcha_ClientException
{}


class DeathByCaptcha_InvalidCaptchaException extends DeathByCaptcha_ClientException
{}


/**
 * Death by Captcha API Client
 *
 * @property-read float|false $balance User's balance (in US cents)
 * @property-read int $status Last API call status (error) code
 * @property-read int $errno Alias for {@link ::$status}
 * @property-read int $error Alias for {@link ::$status}
 *
 * @package DBCAPI
 * @subpackage PHP
 */
class DeathByCaptcha_Client
{
    const API_VERSION        = 'DBC/PHP v3.2';
    const SOFTWARE_VENDOR_ID = 0;

    const API_SERVER_URL = 'http://api.deathbycaptcha.com/api';

    const API_RESPONSE_JSON  = 'application/json';
    const API_RESPONSE_PLAIN = 'text/plain';

    const MAX_CAPTCHA_FILESIZE = 131072;

    const POLLS_COUNT    = 4;
    const POLLS_PERIOD   = 15;
    const POLLS_INTERVAL = 5;

    const DEFAULT_TIMEOUT = 60;


    protected $_conn = null;
    protected $_server_url = null;
    protected $_userpass = array();
    protected $_response = array();
    protected $_response_type = '';
    protected $_response_parser = null;


    /**
     * Verbosity flag
     *
     * @var bool
     */
    public $is_verbose = false;


    /**
     * Parses URL query encoded responses
     *
     * @param string $s
     * @return array
     */
    static public function parse_plain_response($s)
    {
        parse_str($s, $a);
        return $a;
    }

    /**
     * Parses JSON encoded response
     *
     * @param string $s
     * @return array
     */
    static public function parse_json_response($s)
    {
        return json_decode($s, true);
    }


    /**
     * Dumps a message, variable (serialized) or object string to stderr
     *
     * @param mixed $msg
     */
    protected function _dump($msg)
    {
        if ($this->is_verbose) {
            $m = '__toString';
            @fputs(STDERR, (string)time() . ' ' . (
                (is_object($msg) && method_exists($msg, $m))
                    ? $msg->$m()
                    : (is_array($msg)
                        ? serialize($msg)
                        : (string)$msg)
            ) . "\n");
        }
        return $this;
    }

    /**
     * Makes an API call
     *
     * @param string $cmd API command
     * @param string|array $args API call arguments, essentially
     *                           HTTP POST request fields
     * @param bool $login Flag whether login is required
     * @return array|false API response hash table on success
     * @throws DeathByCaptcha_RuntimeException On network related errors
     * @throws DeathByCaptcha_InvalidAccountException On failed login attempt
     * @throws DeathByCaptcha_ServerException On server failures
     */
    protected function _call($cmd, $args=null, $login=true)
    {
        $this->_response = array();
        if (!is_resource($this->_conn)) {
            $this->init();
        }
        $opts = array(
            CURLOPT_URL     => self::API_SERVER_URL . "/{$cmd}/",
            CURLOPT_REFERER => '',
        );
        if ($login) {
            $add_args = array(
                'version'   => self::API_VERSION,
                'swid'      => self::SOFTWARE_VENDOR_ID,
                'username'  => &$this->_userpass[0],
                'password'  => &$this->_userpass[1],
            );
            if (is_array($args)) {
                $args = array_merge($args, $add_args);
            } else {
                $args .= ($args ? '&' : '') . http_build_query($add_args);
            }
        }
        if (null !== $args) {
            $opts[CURLOPT_POST] = true;
            $opts[CURLOPT_POSTFIELDS] = &$args;
        } else {
            $opts[CURLOPT_HTTPGET] = true;
        }
        curl_setopt_array($this->_conn, $opts);
        $this->_dump("SEND: {$cmd} " . serialize($args));
        $response = curl_exec($this->_conn);
        if (!$response || curl_error($this->_conn)) {
            throw new DeathByCaptcha_RuntimeException(
                'Connection failed with the following cURL error: ' .
                '(' . curl_errno($this->_conn) . ') ' . curl_error($this->_conn)
            );
        } else {
            $status_code = curl_getinfo($this->_conn, CURLINFO_HTTP_CODE);
            if (403 == $status_code) {
                throw new DeathByCaptcha_InvalidAccountException(
                    'Access denied, check your credentials and/or balance'
                );
            } else if (500 <= $status_code) {
                throw new DeathByCaptcha_ServerException(
                    "Unexpected HTTP {$status_code} response received"
                );
            }
        }
        $this->_dump("RECV: {$response}");
        $this->_response =
            call_user_func($this->_response_parser, $response);
        if (!$this->_response) {
            throw new DeathByCaptcha_ServerException(
                'Unparseable response received: {$response}'
            );
        }
        return $this->_response;
    }


    /**
     * Cleans up after session, closes opened cURL connection
     */
    public function close()
    {
        if (is_resource($this->_conn)) {
            curl_close($this->_conn);
            $this->_conn = null;
        }
        return $this;
    }

    /**
     * Initializes the client's properties, sets up cURL connection
     */
    public function init()
    {
        $this->close();
        $this->_conn = curl_init();
        if (!$this->_conn) {
            throw new DeathByCaptcha_RuntimeException(
                'Failed initializing a cURL connection'
            );
        }
        curl_setopt_array($this->_conn, array(
            CURLOPT_TIMEOUT        => self::DEFAULT_TIMEOUT,
            CURLOPT_CONNECTTIMEOUT => (int)(self::DEFAULT_TIMEOUT / 4),
            CURLOPT_HEADER         => false,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_FOLLOWLOCATION => true,
            CURLOPT_AUTOREFERER    => false,
            CURLOPT_HTTPHEADER     => array(
                'Accept: ' . $this->_response_type,
                'Expect: ',
                'User-Agent: ' . self::API_VERSION
            )
        ));
        return $this;
    }

    /**
     * Checks runtime environment
     *
     * @param string $username DBC account username
     * @param string $password DBC account password
     * @throws DeathByCaptcha_RuntimeException When required extensions or functions not found
     * @throws DeathByCaptcha_InvalidAccountException On missing or empty DBC credentials
     */
    public function __construct($username, $password)
    {
        $k = 'curl';
        if (!extension_loaded($k)) {
            throw new DeathByCaptcha_RuntimeException(
                "{$k} extension not found"
            );
        }
        if (function_exists('json_decode')) {
            $this->_response_type = self::API_RESPONSE_JSON;
            $this->_response_parser = array($this, 'parse_json_response');
        } else {
            $this->_response_type = self::API_RESPONSE_PLAIN;
            $this->_response_parser = array($this, 'parse_plain_response');
        }
        if (!$username || !$password) {
            throw new DeathByCaptcha_InvalidAccountException(
                'Missing DBC credentials'
            );
        }
        $this->_userpass = array($username, $password);
        $this->init();
    }

    /**
     * @ignore
     */
    public function __destruct()
    {
        $this->close();
    }

    /**
     * @ignore
     */
    public function __get($key)
    {
        switch ($key) {
        case 'balance':
            return $this->get_balance();

        case 'errno':
        case 'error':
        case 'status':
            return (int)@$this->_response['status'];
        }
    }

    /**
     * Returns user details
     *
     * @return array|false
     */
    public function get_user()
    {
        $this->_call('user');
        return !empty($this->_response['user'])
            ? array(
                'user'    => (int)$this->_responser['user'],
                'balance' => (float)$this->_response['balance'],
              )
            : false;
    }

    /**
     * Returns user's balance (in US cents)
     *
     * @return float|false
     */
    public function get_balance()
    {
        return ($a = $this->get_user())
            ? $a['balance']
            : false;
    }

    /**
     * Uploads a CAPTCHA image
     *
     * @param string $file CAPTCHA image file name
     * @return int|false Uploaded CAPTCHA ID on success
     * @throws DeathByCaptcha_InvalidCaptchaException On invalid CAPTCHA file
     */
    public function upload($file)
    {
        if (!$file || !is_file($file) || !is_readable($file)) {
            throw new DeathByCaptcha_InvalidCaptchaException(
                "CAPTCHA image file {$file} not found or unreadable"
            );
        } else if (!filesize($file)) {
            throw new DeathByCaptcha_InvalidCaptchaException(
                "CAPTCHA image file {$file} is empty"
            );
        } else if (self::MAX_CAPTCHA_FILESIZE <= filesize($file)) {
            throw new DeathByCaptcha_InvalidCaptchaException(
                "CAPTCHA image file {$file} is too big"
            );
        }
        $this->_call('captcha', array('captchafile' => "@{$file}"));
        return ($id = (int)@$this->_response['captcha'])
            ? $id
            : false;
    }

    /**
     * Retrieves a CAPTCHA's text if solved
     *
     * @param int $id CAPTCHA id
     * @return string|false
     */
    public function get_text($id)
    {
        $id = max(0, (int)$id);
        return ($id && $this->_call("captcha/{$id}", null, false) && !empty($this->_response['text']))
            ? $this->_response['text']
            : false;
    }

    /**
     * Reports an incorrectly solved CAPTCHA
     *
     * @param int $id CAPTCHA id
     * @return bool
     */
    public function report($id)
    {
        $id = max(0, (int)$id);
        return
            $id && $this->_call("captcha/{$id}/report") &&
            $this->_response['captcha'] && !(bool)@$this->_response['is_correct'];
    }

    /**
     * Removes previously uploaded yet unsolved CAPTCHA
     *
     * @param int $id CAPTCHA id
     * @return bool
     */
    public function remove($id)
    {
        $id = max(0, (int)$id);
        return
            $id && $this->_call("captcha/{$id}/remove") &&
            !$this->_response['captcha'];
    }

    /**
     * Runs the typical sequence of actions:
     * 1) uploads a CAPTCHA image file;
     * 2) polls for decoded text;
     * 3) removes the uploaded CAPTCHA if not solved.
     *
     * @param string $file CAPTCHA image file name
     * @param int $timeout Optional CAPTCHA solving timeout (in seconds)
     * @return array|false (ID, text) tuple if solved
     */
    public function decode($file, $timeout=self::DEFAULT_TIMEOUT)
    {
        if ($id = $this->upload($file)) {
            $attempt = 0;
            $deadline = time() + max(2 * self::POLLS_PERIOD, (int)$timeout);
            while (($deadline > time()) && !$this->_response['text']) {
                $attempt++;
                sleep((1 == ($attempt % self::POLLS_COUNT))
                    ? self::POLLS_PERIOD
                    : self::POLLS_INTERVAL);
                try {
                    $this->get_text($id);
                } catch (Exception $e) {
                    $this->remove($id);
                    throw $e;
                }
            }
            if ($this->_response['text']) {
                return array($id, $this->_response['text']);
            }
            $this->remove($id);
        }
        return false;
    }
}
